#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab

"""
tests.test
~~~~~~~~~~

Provides scripttests pygogo CLI functionality.
"""

import sys

sys.path.append('../chakula')
import chakula                              # noqa

from difflib import unified_diff            # noqa
from os import path as p                    # noqa
from io import StringIO                     # noqa
from timeit import default_timer as timer   # noqa

from scripttest import TestFileEnvironment  # noqa
import pygogo as gogo                       # noqa

CUR_DIR = p.abspath(p.dirname(__file__))
PARENT_DIR = p.abspath(p.dirname(CUR_DIR))


def main(script, tests, verbose=False, stop=True):
    """ Main method

    Returns 0 on success, 1 on failure
    """
    failures = 0
    logger = gogo.Gogo(__name__, verbose=verbose).logger
    short_script = p.basename(script)
    env = TestFileEnvironment('.scripttest')

    start = timer()

    for pos, test in enumerate(tests):
        num = pos + 1
        opts, arguments, expected = test

        if opts or arguments:
            joined_opts = ' '.join(opts) if opts else ''
            joined_args = '"%s"' % '" "'.join(arguments) if arguments else ''
            command = "%s %s %s" % (script, joined_opts, joined_args)
            args = (short_script, joined_opts, joined_args)
            short_command = "%s %s %s" % args
        else:
            command, short_command = script, short_script

        expect_error = not expected if isinstance(expected, bool) else False
        result = env.run(command, cwd=PARENT_DIR, expect_error=expect_error)
        output = result.stdout

        if isinstance(expected, bool):
            outlines = [str(expected)]
            checklines = [str(result.returncode == 0)]
        elif callable(expected):
            outlines = [str(expected(output))]
            checklines = ['True']
        elif p.isfile(expected):
            outlines = StringIO(output).readlines()

            with open(expected, encoding='utf-8') as f:
                checklines = f.readlines()
        else:
            outlines = StringIO(output).readlines()
            checklines = StringIO(expected).readlines()

        args = [checklines, outlines]
        kwargs = {'fromfile': 'expected', 'tofile': 'got'}
        diffs = ''.join(unified_diff(*args, **kwargs))
        passed = not diffs

        if not passed:
            failures += 1
            msg = "ERROR! Output from test #%i:\n  %s\n" % (num, short_command)
            msg += "doesn't match:\n  %s\n" % expected
            msg += diffs if diffs else ''
        else:
            logger.debug(output)
            msg = 'Scripttest #%i: %s ... ok' % (num, short_command)

        logger.info(msg)

        if stop and failures:
            break

    time = timer() - start
    logger.info('%s' % '-' * 70)
    end = 'FAILED (failures=%i)' % failures if failures else 'OK'
    logger.info('Ran %i scripttests in %0.3fs\n\n%s' % (num, time, end))
    sys.exit(failures)


if __name__ == '__main__':
    script = p.join(PARENT_DIR, 'bin', 'chakula')
    feed = p.join(CUR_DIR, 'feeds', 'jenkins.rss')
    cache_path = p.join(PARENT_DIR, '.scripttest', 'cache')
    cache = '--cache {}'.format(cache_path)

    exp = [
        'pip_python2.6 #1006 (FAILURE)\n',
        'pip_python2.6 #1005 (SUCCESS)\n',
        'pip_python2.6 #1004 (SUCCESS)\n',
        'pip_python2.6 #1003 (SUCCESS)\n',
        'pip_python2.6 #1002 (SUCCESS)\n'
    ]

    args = ['--heading', '--iterations 1', '--initial 1', '-s title', '-s url']
    hfunc = lambda stdout: ('Title' in stdout) and ('Url' in stdout)

    tests = [
        (['--help'], [], True),
        (['--version'], [], 'chakula v%s\n' % chakula.__version__),
        (['--iterations 1', '--initial 3'], [feed], ''.join(exp[:3])),
        (['--iterations 1', '-s title'], [feed], ''.join(exp)),
        (
            ['--reverse', '--iterations 1', '-s title'],
            [feed], ''.join(reversed(exp))),
        (args, [feed], hfunc),
        (['--iterations 1', '--newer "2012/01/04 11:00:00"'], [feed], exp[0]),
        (['--iterations 1', '--newer "2012/01/04 11:00"'], [feed], exp[0]),
        (['--iterations 1', '--newer "2012/01/04"'], [feed], ''.join(exp[:2])),
        (['--iterations 1', '--newer "$#@!@#"'], [feed], False),
        ([cache, '--iterations 1', '-s title'], [feed], ''.join(exp)),
        ([cache, '--iterations 1', '-s title'], [feed], ''),
    ]

    main(script, tests)



#!/usr/bin/env python
# encoding: utf-8

from chakula.formatter import Formatter


def test_placeholder_style_detect():
    f = Formatter('{asdf} {zxcv} {qwerty}')
    assert f.is_newstyle

    f = Formatter('%(asdf)s %(zxcv)s %(qwerty)-20s')
    assert not f.is_newstyle

    f = Formatter('{asdf} %(zxcv)s %(qwerty)s')
    assert not f.is_newstyle

    f = Formatter('{asdf} {zxcv} %(qwerty)s')
    assert f.is_newstyle

    f = Formatter('{asdf} {zxcv} %(qwerty)s %(azerty)s')
    assert not f.is_newstyle



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab

import time
from re import findall
from datetime import datetime as dt
from functools import reduce

DEF_TIME_FMT = '%Y/%m/%d %H:%M:%S'


def attrgetter(item, default=''):
    """operator.attrgetter with a default value."""
    reducer = lambda obj, name: getattr(obj, name, default)
    return lambda obj: reduce(reducer, item.split('.'), obj)


PLACEHOLDERS = {
    'author': attrgetter('author'),
    'id': attrgetter('id'),
    'title': attrgetter('title'),
    'link': attrgetter('link'),
    'url': attrgetter('link'),
    'description': attrgetter('description'),
    'desc': attrgetter('description'),
    'pubdate': attrgetter('published_parsed'),
    'updated': attrgetter('updated_parsed'),
    'created': attrgetter('created_parsed'),
    'timestamp': lambda _: None,
    'expired': lambda _: None,
    'comments': lambda _: None,
}


class Formatter(object):
    """I interpolate a format string with feedparser values."""
    def __init__(self, fmt, time_fmt=DEF_TIME_FMT):
        self.fmt = fmt
        self.time_fmt = time_fmt

    @classmethod
    def from_fields(cls, fields, time_fmt=DEF_TIME_FMT, show_heading=False):
        fmt = []

        for field in fields:
            new_line = '\n' if field in {'desc', 'description'} else ''

            if show_heading and field != 'timestamp':
                entry = '{new_line}{tc_field}: %({field})s'
            else:
                entry = '{new_line}%({field})s'

            opts = {
                'field': field, 'tc_field': field.title(),
                'new_line': new_line}

            fmt.append(entry.format(**opts))

        fmt = '{}\n'.format('  '.join(fmt))
        return cls(fmt, time_fmt=time_fmt)

    @property
    def is_newstyle(self):
        """Check if we're dealing with {} or %()s placeholders."""
        new_style = len(findall(r'{[^}]*}', self.fmt))
        old_style = len(findall(r'%\([^\(]*\)[^ ]*s', self.fmt))
        return new_style > old_style

    def __call__(self, entry):
        opts = {field: self.get_value(field, entry) for field in PLACEHOLDERS}
        return self.fmt.format(**opts) if self.is_newstyle else self.fmt % opts

    def get_value(self, field, entry):
        value = PLACEHOLDERS[field](entry)

        if field in {'pubdate', 'updated'}:
            value = time.strftime(self.time_fmt, value)
        elif field == 'timestamp':
            value = dt.utcnow().strftime(self.time_fmt)

        return value



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab

""" A Python logging library with super powers """

import sys
import textwrap

from os import getcwd, path as p
from argparse import RawTextHelpFormatter, ArgumentParser
from pickle import dump, load
from io import open
from functools import partial, lru_cache
from signal import signal, SIGINT

import pygogo as gogo

from dateutil.parser import parse as parse_date
from chakula import tail, __version__
from chakula.formatter import PLACEHOLDERS, Formatter

try:
    from redisworks import Root as OldRoot
except ImportError:
    OldRoot = object


DEF_TIME_FMT = '%Y/%m/%d %H:%M:%S'
DEF_INTERVAL = '300s'
CURDIR = p.basename(getcwd())
LOGFILE = '%s.log' % CURDIR
FIELDS = sorted(PLACEHOLDERS)

logger = gogo.Gogo(__name__, monolog=True).logger

examples = r'''
Format specifiers must have one the following forms:
  %%(placeholder)[flags]s
  {placeholder:flags}

Examples:
  %(prog)s <url>
  echo '<url>' | %(prog)s --reverse
  %(prog)s -s pubdate -s title -s author <url1> <url2> <url3>
  %(prog)s --interval 60s --newer "2011/12/20 23:50:12" <url>
  %(prog)s --format '%%(timestamp)-30s %%(title)s\n' <url>
  %(prog)s --format '%%(title)s was written on %%(pubdate)s\n' <url>
  %(prog)s --format '{timestamp:<30} {title} {author}\n' <url>
  %(prog)s --format '{timestamp:<20} {pubdate:^30} {author:>30}\n' <url>
  %(prog)s --time-format '%%Y/%%m/%%d %%H:%%M:%%S' <url>
  %(prog)s --time-format 'Day of the year: %%j Month: %%b' <url>

Useful flags in this context are:
  %%(placeholder)-10s - left align and pad
  %%(placeholder)10s  - right align and pad
  {placeholder:<10}  - left align and pad
  {placeholder:>10}  - right align and pad
  {placeholder:^10}  - center align and pad
'''

available = textwrap.wrap('Available fields: {}'.format(', '.join(FIELDS)))
epilog = [textwrap.dedent(examples)] + available


def timespec(value):
    """Parse the 'timespec' option:

    >>> timespec(1)
    1
    >>> timespec('5m')
    300
    >>> timespec('1h')
    3600
    """

    try:
        return int(value)
    except ValueError:
        multiply = {'s': 1, 'm': 60, 'h': 3600}
        suffix = value[-1]

        msg = 'invalid timespec value {} - hint: 60, 60s, 1m, 1h'

        if suffix in multiply:
            try:
                v = int(value[:-1])
                return v * multiply[suffix]
            except ValueError:
                ValueError(msg.format(value))
        else:
            raise ValueError(msg.format(value))


parser = ArgumentParser(
    description='description: Tails 1 or more rss feeds',
    prog='chakula',
    usage='%(prog)s [options] <url> [<url> ...]',
    formatter_class=RawTextHelpFormatter,
    epilog='\n'.join(epilog),
)

parser.add_argument(
    dest='urls', nargs='*', default=[sys.stdin],
    help='The urls to tail (default: reads from stdin).')

i_help = 'Number of seconds between polling (default: {}).'

parser.add_argument(
    '-i', '--interval', action='store', help=i_help.format(DEF_INTERVAL),
    type=timespec, default=DEF_INTERVAL)

parser.add_argument(
    '-N', '--iterations', action='store', type=int,
    help='Number of times to poll before quiting (default: inf).')

parser.add_argument(
    '-I', '--initial', action='store', type=int,
    help='Number of entries to show (default: all)')

parser.add_argument(
    '-n', '--newer', metavar='DATE', action='store',
    help='Date by which entries should be newer than')

parser.add_argument(
    '-s', '--show', metavar='FIELD', choices=FIELDS, action='append',
    help='Entry field to display (default: title).', default=[])

t_help = "The date/time format (default: 'YYYY/MM/DD HH:MM:SS')."

parser.add_argument(
    '-t', '--time-format', metavar='FORMAT', action='store',
    default=DEF_TIME_FMT, help=t_help)

parser.add_argument(
    '-F', '--format', action='store',
    help='The output format (overrides other format options).')

parser.add_argument(
    '-c', '--cache', action='store',
    help='File path to store feed information across multiple runs.')

parser.add_argument(
    '-r', '--reverse', action='store_true',
    help='Show entries in reverse order.')

parser.add_argument(
    '-f', '--fail', action='store_true', help='Exit on error.')

parser.add_argument(
    '-u', '--unique', action='store_true', help='Skip duplicate entries.')

parser.add_argument(
    '-H', '--heading', action='store_true', help='Show field headings.')

parser.add_argument(
    '-v', '--version', help="Show version and exit.", action='store_true',
    default=False)

parser.add_argument(
    '-V', '--verbose', help='Increase output verbosity.', action='store_true',
    default=False)


class Root(OldRoot):
    def __init__(self, conn, return_object=True, *args, **kwargs):
        super(Root, self).__init__(*args, **kwargs)
        self.red = conn
        self.return_object = return_object
        self.setup()

get_root = lru_cache(maxsize=8)(lambda conn: Root(conn))


def sigint_handler(signal=None, frame=None):
    logger.info('\nquitting...\n')
    sys.exit(0)


def update_cache(path, extra, redis=False):
    if redis:
        root = get_root(path)

        try:
            items = extra.__dict__['_registry'].evaluated_items
        except AttributeError:
            root.extra = extra
        else:
            root.extra = items['root.extra']

        return root.red
    else:
        with open(path, 'wb') as f:
            dump(extra, f)

        return path


def load_extra(path, redis=False):
    if redis:
        root = get_root(path)
        extra = root.extra or {}

        for k, v in extra.items():
            v['updated'] = tuple(v.get('updated') or [])
            v['modified'] = tuple(v.get('modified') or [])
    else:
        try:
            with open(path, 'rb') as f:
                extra = load(f)
        except FileNotFoundError:
            extra = {}

    return extra


def run():
    """CLI runner"""
    args = parser.parse_args()
    kwargs = {'monolog': True, 'verbose': args.verbose}
    logger = gogo.Gogo(__name__, **kwargs).get_logger('run')
    signal(SIGINT, sigint_handler)

    if args.version:
        logger.info('chakula v%s' % __version__)
        exit(0)

    if args.newer:
        newer = parse_date(args.newer).timetuple()
        logger.debug('showing entries newer than %s', newer)
    else:
        newer = None

    if args.format:
        fmt = args.format.replace('\\n', '\n')
        formatter = Formatter(fmt, args.time_format)
    else:
        show = args.show or ['title']
        pargs = (show, args.time_format, args.heading)
        formatter = Formatter.from_fields(*pargs)

    logger.debug('using format: %r', formatter.fmt)
    logger.debug('using time format: %r', formatter.time_fmt)

    info = {
        'seen': set() if args.unique else None, 'newer': newer,
        'reverse': args.reverse, 'iterations': args.iterations,
        'interval': args.interval, 'formatter': formatter,
        'initial': args.initial, 'logger': logger, 'fail': args.fail}

    first = args.urls[0]

    if hasattr(first, 'isatty') and first.isatty():  # called with no args
        # This doesn't work for scripttest though
        parser.print_help()
        sys.exit(0)
    elif hasattr(first, 'read'):  # piped into sdtin
        urls = first.read().splitlines()
    else:
        urls = args.urls

    if args.cache:
        extra = load_extra(args.cache)
        info['tail_handler'] = partial(update_cache, args.cache)
    else:
        extra = {}

    tail(urls, extra=extra, **info)
    sys.exit(0)


if __name__ == '__main__':
    run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab

import sys

from os import path as p

import pkutils

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

PARENT_DIR = p.abspath(p.dirname(__file__))

sys.dont_write_bytecode = True
requirements = set(pkutils.parse_requirements('requirements.txt'))
dev_requirements = set(pkutils.parse_requirements('dev-requirements.txt'))
readme = pkutils.read('README.rst')
module = pkutils.parse_module(p.join(PARENT_DIR, 'chakula', '__init__.py'))
license = module.__license__
version = module.__version__
project = module.__package_name__
description = module.__description__
user = 'reubano'

setup_require = [r for r in dev_requirements if 'pkutils' in r]

classifiers = [
    pkutils.LICENSES[license],
    pkutils.get_status(version),
    'Natural Language :: English',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: Implementation :: PyPy',
    'Environment :: Console',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Utilities',
    'Framework :: Flask',
    'Operating System :: OS Independent',
    'Operating System :: POSIX :: Linux',
    'Operating System :: MacOS :: MacOS X',
    'Operating System :: Microsoft :: Windows',
]

kw = {
    'name': project,
    'version': version,
    'description': description,
    'long_description': readme,
    'author': module.__author__,
    'author_email': module.__email__,
    'license': license,
    'keywords': [project] + description.split(' '),
    'url': pkutils.get_url(project, user),
    'download_url': pkutils.get_dl_url(project, user, version),
    'classifiers': classifiers,
    'packages': find_packages(exclude=['tests', 'docs']),
    'include_package_data': True,
    'install_requires': requirements,
    'extras_require': {
        'develop': dev_requirements,
        'redis': ['redisworks>=0.2.7,<0.3.0'],
    },
    'setup_requires': setup_require,
    'tests_require': dev_requirements,
    'package_data': {
        'helpers': ['helpers/*'],
        'tests': ['tests/*'],
    },
    'data_files': [],
    'platforms': ['MacOS X', 'Windows', 'Linux'],
    'zip_safe': False,
    'scripts': [p.join('bin', 'chakula')],
}


if __name__ == '__main__':
    setup(**kw)



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab

import sys
import time

from datetime import datetime as dt
from functools import reduce, lru_cache
from traceback import format_exception
from bisect import bisect

import feedparser
import pygogo as gogo

__version__ = '0.8.0'
__title__ = 'chakula'
__package_name__ = 'chakula'
__author__ = 'Reuben Cummings'
__description__ = 'An RSS feed monitor mimicking tail -f'
__email__ = 'reubano@gmail.com'
__license__ = 'BSD'
__copyright__ = 'Copyright 2017 Reuben Cummings'

LOGGER = gogo.Gogo(__name__, monolog=True).logger


@lru_cache(maxsize=32)
def last_update(entry_dates):
    return max(entry_dates) if entry_dates else time.strptime('1900', '%Y')


def write_entries(entries, **kwargs):
    logger = kwargs.get('logger', LOGGER)
    stream = kwargs.get('stream', sys.stdout)
    seen = kwargs.get('seen')
    formatter = kwargs.get('formatter')

    if kwargs.get('reverse'):
        entries.reverse()

    if seen is not None:
        to_add = [entry for entry in entries if entry.id not in seen]
        seen.update(entry.id for entry in to_add)
    else:
        to_add = entries

    logger.debug('Writing {} new entries'.format(len(to_add)))

    for entry in to_add:
        if formatter:
            content = formatter(entry)
        else:
            content = '{}\n'.format(entry['title'])

        try:
            stream.write(content)
        except TypeError:
            stream.write(content.encode('utf-8'))

    try:
        stream.flush()
    except AttributeError:
        pass

    if kwargs.get('write_handler'):
        kwargs['write_handler'](to_add)


def parse_url(url, iteration, initial=None, **kwargs):
    logger = kwargs.get('logger', LOGGER)
    updated = kwargs.get('updated')
    newer = kwargs.get('newer')
    pkwargs = {k: v for k, v in kwargs.items() if k in {'etag', 'modified'}}
    feed = feedparser.parse(url, **pkwargs)

    if feed.bozo == 1:
        safeexc = (feedparser.CharacterEncodingOverride,)

        if not isinstance(feed.bozo_exception, safeexc):
            msg = 'feed error %r:\n%s' % (url, feed.bozo_exception)
            raise ValueError(msg)

    entries = feed.entries if iteration else feed.entries[:initial]

    if updated and newer:
        newer_than = max([updated, newer])
    else:
        newer_than = updated or newer

    if newer_than:
        formatted = time.strftime('%Y/%m/%d %H:%M:%S', newer_than)
        logger.debug('selecting entries newer than %s', formatted)
        entries = [
            entry for entry in entries if entry.published_parsed > newer_than]

    if not feed.get('updated_parsed') and entries:
        dates = (e.updated_parsed for e in entries if e.get('updated_parsed'))
        def_updated = last_update(tuple(dates))
    else:
        def_updated = updated

    info = {
        'etag': feed.get('etag', kwargs.get('etag')),
        'modified': feed.get('modified_parsed'),
        'updated': feed.get('updated_parsed') or def_updated}

    return entries, info


def parse_interval(interval):
    pairs = [('secs', 0), ('minutes', 60), ('hours', 3600), ('days', 86400)]
    grades = [pair[0] for pair in pairs]
    breakpoints = [pair[1] for pair in pairs][1:]

    def get_grade(score):
        i = bisect(breakpoints, score)
        return grades[i]

    grade = get_grade(interval)
    divisor = dict(pairs)[grade]
    return round(interval / divisor, 2), grade


def tail(urls, iteration=0, interval=300, extra=None, **kwargs):
    logger = kwargs.get('logger', LOGGER)
    extra = extra or {}

    if kwargs.get('iterations') and iteration >= kwargs['iterations']:
        msg = 'maximum number of iterations reached: %d'
        logger.info(msg, kwargs['iterations'])
        return extra
    elif iteration:
        # sleep first so that we don't have to wait an interval before checking
        # iteration count
        parsed = parse_interval(interval)
        logger.info('sleeping for {} {}'.format(*parsed))
        time.sleep(interval)

    for url in urls:
        kwargs.update(extra.get(url, {}))

        try:
            entries, extra[url] = parse_url(url, iteration, **kwargs)
            kwargs.update(extra[url])
            write_entries(entries, **kwargs)
        except Exception:
            if kwargs.get('fail'):
                raise
            else:
                traceback_strings = format_exception(*sys.exc_info())
                logger.error(''.join(traceback_strings))

    if kwargs.get('tail_handler'):
        kwargs['tail_handler'](extra)

    return tail(urls, iteration + 1, interval=interval, extra=extra, **kwargs)
